import numpy as np
from PIL.Image import Image as PIL_Image

DEFAULT_RATIO = np.sqrt(49 / 400)

def __expand_parts(p):
	return p[:, 0], p[:, 1:3], p[:, 3].astype(bool)

def dimensions(im):
	if isinstance(im, np.ndarray):
		assert im.ndim == 3, "Only RGB images are currently supported!"
		return im.shape
	elif isinstance(im, PIL_Image):
		w, h = im.size
		c = len(im.getbands())
		# assert c == 3, "Only RGB images are currently supported!"
		return h, w, c
	else:
		raise ValueError("Unknown image instance ({})!".format(type(im)))

def asarray(im, dtype=np.uint8):
	if isinstance(im, np.ndarray):
		return im.astype(dtype)
	elif isinstance(im, PIL_Image):
		return np.asarray(im, dtype=dtype)
	else:
		raise ValueError("Unknown image instance ({})!".format(type(im)))


def uniform_parts(im, ratio=DEFAULT_RATIO, round_op=np.floor):
	h, w, c = dimensions(im)

	part_w = round_op(w * ratio).astype(np.int32)
	part_h = round_op(h * ratio).astype(np.int32)

	n, m = w // part_w, h // part_h

	parts = np.ones((n*m, 4), dtype=int)
	parts[:, 0] = np.arange(n*m)

	for x in range(n):
		for y in range(m):
			i = y * n + x
			x0, y0 = x * part_w, y * part_h
			parts[i, 1:3] = [x0 + part_w // 2, y0 + part_h // 2]

	return parts

def visible_part_locs(p):
	idxs, locs, vis = __expand_parts(p)
	return idxs[vis], locs[vis].T


def crops(im, xy, ratio=DEFAULT_RATIO, padding_mode="edge"):
	h, w, c = dimensions(im)
	crop_h, crop_w = int(h * ratio), int(w * ratio)
	crops = np.zeros((xy.shape[1], crop_h, crop_w, c), dtype=np.uint8)

	pad_h, pad_w = crop_h // 2, crop_w // 2

	padded_im = np.pad(im, [(pad_h, pad_h), (pad_w, pad_w), [0,0]], mode=padding_mode)

	for i, (x, y) in enumerate(xy.T):
		x0, y0 = x - crop_w // 2 + pad_w, y - crop_h // 2 + pad_h
		crops[i] = padded_im[y0:y0+crop_h, x0:x0+crop_w]

	return crops

def visible_crops(im, p, *args, **kw):
	idxs, locs, vis = __expand_parts(p)
	parts = crops(asarray(im), locs[vis].T, *args, **kw)
	res = np.zeros((len(idxs),) + parts.shape[1:], dtype=parts.dtype)
	res[vis] = parts
	return res

def reveal_parts(im, xy, ratio=DEFAULT_RATIO):
	h, w, c = dimensions(im)
	crop_h, crop_w = int(h * ratio), int(w * ratio)
	im = asarray(im)
	res = np.zeros_like(im)
	for x, y in xy.T:
		x0, y0 = max(x - crop_w // 2, 0), max(y - crop_h // 2, 0)
		res[y0:y0+crop_h, x0:x0+crop_w] = im[y0:y0+crop_h, x0:x0+crop_w]

	return res

def select(crops, mask):
	selected = np.zeros_like(crops)
	selected[mask] = crops[mask]
	return selected

def selection_mask(idxs, n):
	return np.bincount(idxs, minlength=n).astype(bool)

def random_select(idxs, xy, part_crops, *args, **kw):
	rnd_idxs = random_idxs(np.arange(len(idxs)), *args, **kw)
	idxs = idxs[rnd_idxs]
	xy = xy[:, rnd_idxs]

	mask = selection_mask(idxs, len(part_crops))
	selected_crops = select(part_crops, mask)

	return idxs, xy, selected_crops

def random_idxs(idxs, rnd=None, n_parts=None):

	if rnd is None or isinstance(rnd, int):
		rnd = np.random.RandomState(rnd)
	else:
		assert isinstance(rnd, np.random.RandomState), \
			"'rnd' should be either a random seed or a RandomState instance!"

	n_parts = n_parts or rnd.randint(1, len(idxs))
	res = rnd.choice(idxs, n_parts, replace=False)
	res.sort()
	return res



import numpy as np

from . import BaseMixin

class BBCropMixin(BaseMixin):

	def __init__(self, crop_to_bb=False, crop_uniform=False, *args, **kwargs):
		super(BBCropMixin, self).__init__(*args, **kwargs)
		self.crop_to_bb = crop_to_bb
		self.crop_uniform = crop_uniform

	def bounding_box(self, i):
		bbox = self._get("bounding_box", i)
		x,y,w,h = [bbox[attr] for attr in "xywh"]
		if self.crop_uniform:
			x0 = x + w//2
			y0 = y + h//2

			crop_size = max(w//2, h//2)

			x,y = max(x0 - crop_size, 0), max(y0 - crop_size, 0)
			w = h = crop_size * 2
		return x,y,w,h

	def get_example(self, i):
		im_obj = super(BBCropMixin, self).get_example(i)
		if self.crop_to_bb:
			bb = self.bounding_box(i)
			return im_obj.crop(*bb)
		return im_obj

class PartsInBBMixin(BaseMixin):
	def __init__(self, parts_in_bb=False, *args, **kwargs):
		super(PartsInBBMixin, self).__init__(*args, **kwargs)
		self.parts_in_bb = parts_in_bb

	def get_example(self, i):
		im_obj = super(PartsInBBMixin, self).get_example(i)

		if self.parts_in_bb:
			bb = self.bounding_box(i)
			return im_obj.hide_parts_outside_bb(*bb)
		return im_obj

class PartCropMixin(BaseMixin):

	def __init__(self, return_part_crops=False, *args, **kwargs):
		super(PartCropMixin, self).__init__(*args, **kwargs)
		self.return_part_crops = return_part_crops

	def get_example(self, i):
		im_obj = super(PartCropMixin, self).get_example(i)
		if self.return_part_crops:
			return im_obj.part_crops(self.ratio)
		return im_obj


class PartRevealMixin(BaseMixin):

	def __init__(self, reveal_visible=False, *args, **kwargs):
		super(PartRevealMixin, self).__init__(*args, **kwargs)
		self.reveal_visible = reveal_visible

	def get_example(self, i):
		im_obj = super(PartRevealMixin, self).get_example(i)
		assert hasattr(self, "ratio"), "\"ratio\" attribute is missing!"
		if not self.reveal_visible:
			return im_obj.reveal_visible(self.ratio)
		return im_obj


class UniformPartMixin(BaseMixin):

	def __init__(self, uniform_parts=False, ratio=None, *args, **kwargs):
		super(UniformPartMixin, self).__init__(*args, **kwargs)
		self.uniform_parts = uniform_parts
		self.ratio = ratio

	def get_example(self, i):
		im_obj = super(UniformPartMixin, self).get_example(i)
		if self.uniform_parts:
			return im_obj.uniform_parts(self.ratio)
		return im_obj

class RandomBlackOutMixin(BaseMixin):

	def __init__(self, seed=None, rnd_select=False, n_parts=None, *args, **kwargs):
		super(RandomBlackOutMixin, self).__init__(*args, **kwargs)
		self.rnd = np.random.RandomState(seed)
		self.rnd_select = rnd_select
		self.n_parts = n_parts

	def get_example(self, i):
		im_obj = super(RandomBlackOutMixin, self).get_example(i)
		if self.rnd_select:
			return im_obj.select_random_parts(rnd=self.rnd, n_parts=self.n_parts)
		return im_obj


# some shortcuts

class PartMixin(RandomBlackOutMixin, PartsInBBMixin, UniformPartMixin, BBCropMixin):
	"""
		TODO!
	"""

class RevealedPartMixin(PartRevealMixin, PartMixin):
	"""
		TODO!
	"""


class CroppedPartMixin(PartCropMixin, PartMixin):
	"""
		TODO!
	"""



#!/usr/bin/env python
if __name__ != '__main__': raise Exception("Do not import me!")

from argparse import ArgumentParser
import logging
import numpy as np

from annotations import NAB_Annotations, CUB_Annotations
from dataset import Dataset
from dataset.utils import reveal_parts, uniform_parts, \
	random_select, \
	visible_part_locs, visible_crops


import matplotlib.pyplot as plt

def init_logger(args):
	fmt = "%(levelname)s - [%(asctime)s] %(filename)s:%(lineno)d [%(funcName)s]: %(message)s"
	logging.basicConfig(
		format=fmt,
		level=getattr(logging, args.loglevel.upper(), logging.DEBUG),
		filename=args.logfile or None,
		filemode="w")

def plot_crops(crops, title, scatter_mid=False):

	fig = plt.figure(figsize=(16,9))
	fig.suptitle(title, fontsize=16)

	n_crops = crops.shape[0]
	rows = int(np.ceil(np.sqrt(n_crops)))
	cols = int(np.ceil(n_crops / rows))

	for j, crop in enumerate(crops, 1):
		ax = fig.add_subplot(rows, cols, j)
		ax.imshow(crop)
		ax.axis("off")
		if scatter_mid:
			middle_h, middle_w = crop.shape[0] / 2, crop.shape[1] / 2
			ax.scatter(middle_w, middle_h, marker="x")



def main(args):
	init_logger(args)

	annotation_cls = dict(
		nab=NAB_Annotations,
		cub=CUB_Annotations)

	logging.info("Loading \"{}\" annnotations from \"{}\"".format(args.dataset, args.data))
	annot = annotation_cls.get(args.dataset.lower())(args.data)

	subset = args.subset.lower()

	uuids = getattr(annot, "{}_uuids".format(subset))
	features = args.features[0 if subset == "train" else 1]

	data = Dataset(
		uuids=uuids, annotations=annot,
		features=features,

		uniform_parts=args.uniform_parts,

		crop_to_bb=args.crop_to_bb,
		crop_uniform=args.crop_uniform,

		parts_in_bb=args.parts_in_bb,

		rnd_select=args.rnd,
		ratio=args.ratio,
		seed=args.seed

	)
	n_images = len(data)
	logging.info("Found {} images in the {} subset".format(n_images, subset))

	for i in range(n_images):
		if i + 1 <= args.start: continue
		im, parts, label = data[i]

		idxs, xy = visible_part_locs(parts)
		part_crops = visible_crops(im, parts, ratio=args.ratio)
		if args.rnd:
			selected = parts[:, -1].astype(bool)
			parts[selected, -1] = 0
			parts[np.logical_not(selected), -1] = 1
			action_crops = visible_crops(im, parts, ratio=args.ratio)

		logging.debug(label)
		logging.debug(idxs)
		logging.debug(xy)

		fig1 = plt.figure(figsize=(16,9))
		ax = fig1.add_subplot(2,1,1)
		ax.imshow(im)
		ax.set_title("Visible Parts")
		ax.scatter(*xy, marker="x", c=idxs)
		ax.axis("off")

		ax = fig1.add_subplot(2,1,2)
		ax.set_title("{}selected parts".format("randomly " if args.rnd else ""))
		ax.imshow(reveal_parts(im, xy, ratio=args.ratio))
		# ax.scatter(*xy, marker="x", c=idxs)
		ax.axis("off")

		plot_crops(part_crops, "Selected parts")

		if args.rnd:
			plot_crops(action_crops, "Actions")

		plt.show()
		plt.close()

		if i+1 >= args.start + args.n_images: break

parser = ArgumentParser()

parser.add_argument("data",
	help="Folder containing the dataset with images and annotation files",
	type=str)

parser.add_argument("--dataset",
	help="Possible datasets: NAB, CUB",
	choices=["cub", "nab"],
	default="nab", type=str)

parser.add_argument("--features",
	help="pre-extracted train and test features",
	default=[None, None],
	nargs=2, type=str)

parser.add_argument("--subset",
	help="Possible subsets: train, test",
	choices=["train", "test"],
	default="train", type=str)

parser.add_argument("--start", "-s",
	help="Image id to start with",
	type=int, default=0)

parser.add_argument("--n_images", "-n",
	help="Number of images to display",
	type=int, default=10)

parser.add_argument("--ratio",
	help="Part extraction ratio",
	type=float, default=.2)

parser.add_argument("--rnd",
	help="select random subset of present parts",
	action="store_true")

parser.add_argument("--uniform_parts", "-u",
	help="Do not use GT parts, but sample parts uniformly from the image",
	action="store_true")

parser.add_argument("--crop_to_bb",
	help="Crop image to the bounding box",
	action="store_true")

parser.add_argument("--crop_uniform",
	help="Try to extend the bounding box to same height and width",
	action="store_true")

parser.add_argument("--parts_in_bb",
	help="Only display parts, that are inside the bounding box",
	action="store_true")



parser.add_argument(
	'--logfile', type=str, default='',
	help='File for logging output')

parser.add_argument(
	'--loglevel', type=str, default='INFO',
	help='logging level. see logging module for more information')

parser.add_argument(
	'--seed', type=int, default=12311123,
	help='random seed')

main(parser.parse_args())



import numpy as np

from os.path import isfile

from . import BaseMixin


class PreExtractedFeaturesMixin(BaseMixin):

	def __size_check(self):
		assert len(self.features) == len(self), \
			"Number of features ({}) does not match the number of images ({})!".format(
				len(self.features), len(self)
			)

	def __init__(self, features=None, *args, **kw):
		super(PreExtractedFeaturesMixin, self).__init__(*args, **kw)

		self.features = None
		if features is not None and isfile(features):
			self.features = self.load_features(features)
			self.__size_check()

	def load_features(self, features_file):
		"""
			Default feature loading from a file.
			If you desire another feature loading logic,
			subclass this mixin and override this method.
		"""
		try:
			cont = np.load(features_file)
			return cont["features"]
		except Exception as e:
			msg = "Error occured while reading features: \"{}\". ".format(e) + \
				"If you want another feature loading logic, override this method!"
			raise ValueError(msg)

	def get_example(self, i):
		im_obj = super(PreExtractedFeaturesMixin, self).get_example(i)
		if self.features is not None:
			im_obj.feature = self.features[i]

		return im_obj



from imageio import imread
from PIL import Image
from os.path import isfile

import copy
import numpy as np

from . import utils

def should_have_parts(func):
	def inner(self, *args, **kwargs):
		assert self.has_parts, "parts are not present!"
		return func(self, *args, **kwargs)
	return inner

class ImageWrapper(object):
	@staticmethod
	def read_image(im_path, mode="RGB"):
		# im = imread(im_path, pilmode=mode)
		im = Image.open(im_path, mode="r")
		return im


	def __init__(self, im_path, label, parts=None, mode="RGB"):

		self.mode = mode
		self.im = im_path
		self._im_array = None

		self.label = label
		self.parts = parts

		self.parent = None
		self._feature = None

	def __del__(self):
		if isinstance(self._im, Image.Image):
			if self._im is not None and self._im.fp is not None:
				self._im.close()

	@property
	def im(self):
		return self._im

	@property
	def im_array(self):
		if self._im_array is None:
			if isinstance(self._im, Image.Image):
				_im = self._im.convert(self.mode)
				self._im_array = utils.asarray(_im)
			elif isinstance(self._im, np.ndarray):
				if self.mode == "RGB" and self._im.ndim == 2:
					self._im_array = np.stack((self._im,) * 3, axis=-1)
				elif self._im.ndim == 3:
					self._im_array = self._im
				else:
					raise ValueError()
			else:
				raise ValueError()
		return self._im_array

	@im.setter
	def im(self, value):
		if isinstance(value, str):
			assert isfile(value), "Image \"{}\" does not exist!".format(value)
			self._im = ImageWrapper.read_image(value, mode=self.mode)
		else:
			self._im = value

	def as_tuple(self):
		return self.im_array, self.parts, self.label

	def copy(self):
		new = copy.copy(self)
		new.parent = self
		deepcopies = [
			"_feature",
			"parts",
		]
		for attr_name in deepcopies:
			attr_copy = copy.deepcopy(getattr(self, attr_name))
			setattr(new, attr_name, attr_copy)

		return new
	@property
	def feature(self):
		return self._feature

	@feature.setter
	def feature(self, im_feature):
		self._feature = im_feature

	def crop(self, x, y, w, h):
		result = self.copy()
		# result.im = self.im[y:y+h, x:x+w]
		result.im = self.im.crop(x, y, x+w, y+h)
		if self.has_parts:
			result.parts[:, 1] -= x
			result.parts[:, 2] -= y
		return result

	@should_have_parts
	def hide_parts_outside_bb(self, x, y, w, h):
		idxs, (xs,ys) = self.visible_part_locs()
		f = np.logical_and
		mask = f(f(x <= xs, xs <= x+w), f(y <= ys, ys <= y+h))
		result = self.copy()
		result.parts[:, -1] = mask.astype(self.parts.dtype)

		return result

	def uniform_parts(self, ratio):
		result = self.copy()
		result.parts = utils.uniform_parts(self.im, ratio=ratio)
		return result

	@should_have_parts
	def select_parts(self, idxs):
		result = self.copy()

		result.parts[:, -1] = 0
		result.parts[idxs, -1] = 1

		return result

	@should_have_parts
	def select_random_parts(self, rnd, n_parts):

		idxs, xy = self.visible_part_locs()
		rnd_idxs = utils.random_idxs(idxs, rnd=rnd, n_parts=n_parts)
		return self.select_parts(rnd_idxs)

	@should_have_parts
	def visible_crops(self, ratio):
		return utils.visible_crops(self.im, self.parts, ratio=ratio)

	@should_have_parts
	def visible_part_locs(self):
		return utils.visible_part_locs(self.parts)

	@should_have_parts
	def reveal_visible(self, ratio):
		_, xy = self.visible_part_locs()
		result = self.copy()
		result.im = utils.reveal_parts(self.im, xy, ratio=ratio)
		return result

	@should_have_parts
	def part_crops(self, ratio):
		crops = self.visible_crops(ratio)
		idxs, _ = self.visible_part_locs()
		result = self.copy()
		result.im = crops[idxs]
		return result

	@property
	def has_parts(self):
		return self.parts is not None




from os.path import join

from . import BaseMixin
from ..image import ImageWrapper

class AnnotationsReadMixin(BaseMixin):

	def __init__(self, uuids, annotations, mode="RGB"):
		super(AnnotationsReadMixin, self).__init__()
		self.uuids = uuids
		self._annot = annotations
		self.mode = mode

	def __len__(self):
		return len(self.uuids)

	def _get(self, method, i):
		return getattr(self._annot, method)(self.uuids[i])

	def get_example(self, i):
		res = super(AnnotationsReadMixin, self).get_example(i)
		# if the super class returns something, then the class inheritance is wrong
		assert res is None, "AnnotationsReadMixin should be the last class in the hierarchy!"

		methods = ["image", "parts", "label"]
		im_path, parts, label = [self._get(m, i) for m in methods]

		return ImageWrapper(im_path, int(label), parts, mode=self.mode)

	@property
	def labels(self):
		return np.array([self._get("label", i) for i in range(len(self))])


class ImageListReadingMixin(BaseMixin):

	def __init__(self, pairs, root="."):
		super(ImageListReadingMixin, self).__init__()
		with open(pairs) as f:
			self._pairs = [line.strip().split() for line in f]

		assert all([len(pair) == 2 for pair in self._pairs]), \
			"Invalid format of the pairs file!"

		self._root = root

	def __len__(self):
		return len(self._pairs)

	def get_example(self, i):
		im_file, label = self._pairs[i]
		im_path = join(self._root, im_file)

		return ImageWrapper(im_path, int(label))

	@property
	def labels(self):
		return np.array([label for (_, label) in self._pairs])




#!/usr/bin/env python

import os
import pkg_resources
import sys

from setuptools import setup, find_packages

try: # for pip >= 10
	from pip._internal.req import parse_requirements
except ImportError: # for pip <= 9.0.3
	from pip.req import parse_requirements

import nabirds
install_requires = [line.strip() for line in open("requirements.txt").readlines()]

setup(
	name='nabirds',
	version=nabirds.__version__,
	description='Wrapper (inofficial) for NA-Birds bataset (http://dl.allaboutbirds.org/nabirds)',
	author='Dimitri Korsch',
	author_email='korschdima@gmail.com',
	license='MIT License',
	packages=find_packages(),
	zip_safe=False,
	setup_requires=[],
	install_requires=install_requires,
	package_data={'': ['requirements.txt']},
	data_files=[('.',['requirements.txt'])],
	include_package_data=True,
)



from abc import ABC, abstractmethod

class BaseMixin(ABC):

	@abstractmethod
	def get_example(self, i):
		s = super(BaseMixin, self)
		if hasattr(s, "get_example"):
			return s.get_example(i)

	def __getitem__(self, i):
		return self.get_example(i)



from .mixins.reading import AnnotationsReadMixin, ImageListReadingMixin
from .mixins.parts import PartMixin, RevealedPartMixin, CroppedPartMixin
from .mixins.features import PreExtractedFeaturesMixin

class Dataset(PartMixin, PreExtractedFeaturesMixin, AnnotationsReadMixin):

	def get_example(self, i):
		im_obj = super(Dataset, self).get_example(i)
		return im_obj.as_tuple()
